########
#
# function for SImodels with LINear Predictor by BOOSTind with cut version
#
######

# y: response variable
# X: predictor matrix
# sam: index for samples of y and X which are used for estimation
# m.h: number of knots used by the response function
# deg.h: degrees of freedom of the response function
# nu.h: weakness of the learner of the response function
# lambda.h: smoothing parameter for the response functio
# nu.f: weakness of the parameter vector estimator
# M: maximal number of boosting iteration
# family: distributional assumption
# monotone: monotonize the response function?
library(fda)
library(quadprog)  
SI.LIN.BOOST.cut <- function(y,X,sam=j.in,m.h,deg.h,nu.h,lambda.h,nu.f,M,family,monotone="TRUE")
{
### Data standardization ###
X.org<-X
X<-scale(X.org[sam,],center=TRUE,scale=TRUE)
X.out<-scale(X.org[-sam,],center=attr(X,"scaled:center"),scale=attr(X,"scaled:scale"))
X.all<-rbind(X.out,X)
y<-y[sam]
n<-length(y)
p<-ncol(X)
nderiv1<-rep(1,n)

### Initialization of response function ###
n.basis<-m.h-2+deg.h+1
XMAX<-max(sqrt(rowSums(X.all^2)))
knots.h<-seq(-XMAX,XMAX,length=m.h)
B.h <- matrix(0,n,n.basis)
B.h.alt <- matrix(0,n,n.basis)

### initialization of linear predictor ###
eta.iter.norm <- matrix(0,n,M)
eta.iter.alt<- matrix(0,n)
### initialization of basis expansion parameters of the response function ###
alpha.norm <- matrix(0,n.basis,M)
alpha.alt<-rep(0,n.basis)
### initialization of parameter vector of the linear predictor ###
beta <- matrix(0,p,M)
beta.alt<-rep(0,p)
beta.work <- matrix(0,p,p)
beta.cut <- matrix(0,p,M)
### initialization of deviance vector ###
dev.vec <- rep(0,M)
### initialization of fs-matrix ###
FS<-matrix(0,p,p)
### matrix for monotone restriction ###
A<--cbind(diag(m.h+deg.h-2),0)+cbind(0,diag(m.h+deg.h-2))
### initialization of prediction values ###
mu.hat<-matrix(0,n,M)
mu.hat.alt<-matrix(0,n,M)
### initialization of select vector ###
Select<-rep(0,M)
### computing initial values ###
offset=family$linkfun(mean(y))
XMAX<-max(knots.h)
Test2<-create.bspline.basis(knots.h,nbasis=n.basis,norder=deg.h+1)
P.h<-lambda.h*(bsplinepen(Test2))
B.h.alt<-bsplineS(eta.iter.alt,breaks=knots.h,norder=deg.h+1)
Test<-bsplineS(seq(-XMAX,XMAX,length=1000),breaks=knots.h)
alpha.0<-coef(lm(seq(-0.0001*XMAX+offset,0.0001*XMAX+offset,length=1000)~Test-1)) #-log(mean(y)/(1-mean(y)))
alpha.alt<-alpha.0
D.h.im1<- bsplineS(eta.iter.alt,breaks=knots.h,norder=deg.h+1,nderiv=1)%*%alpha.0
D.h0<-family$mu.eta(D.h.im1)
D.h0h<-D.h0*D.h.im1
h.im1<-B.h.alt%*%alpha.0
mu.i<-family$linkinv(h.im1)
S<-family$variance(mu.i)
dev.alpha<-Inf

########
#
# Begin of Boosting Iteration
#
#########
for(i in 1:M)
{
##########
#
#  Fisher-Scoring for Parameter Vector
#
###########
  mu.i.work<-rep(0,n)
  dev.iter<-rep(0,p)
  W1<-D.h0h*S^(-1)
  W2<-(D.h0h)^2 *S^(-1)
  FS<-diag(nu.f*rowSums((t(X^2)%*%diag(c(W2))))^(-1)*rowSums(t(X)%*%diag(c(W1* (y-mu.i)))))

  for(k in 1:p)
  {
    beta.aux<-beta.alt+FS[,k]
    beta.aux.l2<-beta.aux/sqrt(sum(beta.aux^2))
    eta.iter.work.norm<-X%*%beta.aux.l2
    B.h<-bsplineS(eta.iter.work.norm,breaks=knots.h,norder=deg.h+1)
    h.eta.work<-(B.h%*%alpha.alt)
    mu.i.work<-family$linkinv(h.eta.work)
    dev.iter[k]<-sum(family$dev.resids(y,mu.i.work,nderiv1))
    beta.work[,k]<-beta.aux.l2
  }
  r.iter<-which.min(dev.iter)
  dev.beta<-dev.iter[r.iter]
  Select[i]<-r.iter
##########
#
#  Fisher-Scoring for the Response Function
#
###########
  W<-diag(c((D.h0)^2*S^(-1)))
  if(monotone=="FALSE"&i!=1)
  {
    alpha.update<-nu.h*ridge.lmn.short(y=y,X=B.h.alt,fit=mu.i,d=D.h0,P=P.h,S=S)  #+0.000001*diag(m.h+deg.h-1)
    alpha.work<-alpha.alt+alpha.update
  }
  if(monotone=="TRUE"&i!=1)
  {
    H<-t(B.h.alt)%*%W%*%B.h.alt + P.h+0.0001*diag(m.h+deg.h-1)
    y.tilde<-t(B.h.alt)%*%diag(c(D.h0*S^(-1)))%*%(y-mu.i)
    b<--A%*%alpha.alt
    hui<-solve.QP(H,y.tilde, t(A),b)
    alpha.work<-alpha.alt+nu.h*hui$solution
  }
  if(i==1)alpha.work<-alpha.alt
    mu.i.work<-family$linkinv(B.h.alt%*%alpha.work)
    dev.alpha<- sum(family$dev.resids(y,mu.i.work,nderiv1))
##########
#
#  What kind of Update
#
###########
  
  ### Update der Responsefunktion
  if(dev.alpha<=dev.beta)
  {
    beta[,i]<-beta.alt
    b.cut<-beta[,i]    
    a<-abs(b.cut)<1/p
    b.cut[a]<-0
    if(sum(b.cut^2)==0)beta.cut[,i]<-rep(0,p)
    if(sum(b.cut^2)!=0)beta.cut[,i]<-b.cut/sqrt(sum(b.cut^2))

    Select[i]<-0
    eta.iter.norm[,i] <- eta.iter.alt#<-X%*%beta.alt#*const                         
    alpha.norm[,i]<-alpha.alt<-alpha.work
    D.h.im1<-bsplineS(x=eta.iter.alt, breaks=knots.h, norder=deg.h+1, nderiv=1)%*%alpha.alt
    h.alt<-B.h.alt%*%alpha.alt
    D.h0<-family$mu.eta(h.alt)
    D.h0h<-D.h0*D.h.im1
    mu.i<-family$linkinv(h.alt)
    S<-family$variance(mu.i)
    dev.vec[i]<-sum(family$dev.resids(y,mu.i,nderiv1))
     
  }
  
  ### Update des Predictors
  if(dev.alpha>dev.beta)
  {
    alpha.norm[,i]<-alpha.alt
    beta[,i]<-beta.alt<-beta.work[,r.iter]
    b.cut<-beta[,i]    
    a<-abs(b.cut)<1/p
    b.cut[a]<-0
    if(sum(b.cut^2)==0)beta.cut[,i]<-rep(0,p)
    if(sum(b.cut^2)!=0)beta.cut[,i]<-b.cut/sqrt(sum(b.cut^2))


    eta.iter.norm[,i]<-eta.iter.alt<-X%*%beta.alt#*const
    D.h.im1<-bsplineS(x=eta.iter.alt, breaks=knots.h, norder=deg.h+1, nderiv=1)%*%alpha.alt
    B.h.alt<-bsplineS(eta.iter.alt,breaks=knots.h, norder=deg.h+1, nderiv=0)
    h.alt <-B.h.alt%*%alpha.alt
    D.h0<-family$mu.eta(h.alt)
    D.h0h<-D.h0*D.h.im1
    mu.i<-family$linkinv(h.alt)
    S<-family$variance(mu.i)    
    dev.vec[i]<-sum(family$dev.resids(y,mu.i,nderiv1))
    
  }
   mu.hat[,i]<-mu.i
  
 }    
  


  result <- list("Eta.iter.norm"=eta.iter.norm,
                 "Alpha.norm"=alpha.norm,
                 "Beta"= beta,
                 "Beta.cut"= beta.cut,
                 "knots.h"=knots.h,
                 "MU.hat"=mu.hat,
                 "Select"=Select,
                 "deviance.vec"=dev.vec,
                 "Deg.h"=deg.h)

}








########
#
# CV for SI.LIN.BOOST.cut
#
######



CV.SI.LIN.BOOST.cut<-function(y,X,split,m.h=20,deg.h=3,nu.h=0.1,lambda.h=100,nu.f=0.1,M=500,family,monotone="TRUE")
{
n<-length(y)
p<-ncol(X)
dev.vec<-rep(0,M)
dev.vec.cut<-rep(0,M)

for(j in 1:ncol(split.M))
{
  j.out <- which(split.M[,j]==0)
  j.in <- which(split.M[,j]==1)
  CV.TEST<-SI.LIN.BOOST.cut(y,X,sam=j.in,m.h,deg.h,nu.h,lambda.h,nu.f,M,family,monotone="TRUE")
  X.out<-X[j.out,]
  X.in<-X[j.in,]
  X.out<-scale(X.out,center=colMeans(X.in),scale=sd(X.in))
  y.out<-y[j.out]
  a.out<-CV.TEST$Alpha.norm
  b.out<-CV.TEST$Beta
  b.out.cut<-CV.TEST$Beta.cut    
  eta.norm.out<-X.out%*%b.out
  eta.norm.out.cut<-X.out%*%b.out.cut
  n.out<-length(y.out)
  KNOTS<-CV.TEST$knots.h
  for(k in 1:M)
  {
    B.out<-bsplineS(eta.norm.out[,k],breaks=KNOTS, norder=deg.h+1, nderiv=0)
    B.out.cut<-bsplineS(eta.norm.out.cut[,k],breaks=KNOTS, norder=deg.h+1, nderiv=0)
    h.out<-family$linkinv(B.out%*%a.out[,k])
    dev.vec[k]<-sum(family$dev.resids(y.out,h.out,rep(1,n.out)))+dev.vec[k]
    h.out.cut<-family$linkinv(B.out.cut%*%a.out[,k])
    dev.vec.cut[k]<-sum(family$dev.resids(y.out,h.out.cut,rep(1,n.out)))+dev.vec.cut[k]
  }  
}
m.opt<-which.min(dev.vec)
optMOD<-SI.LIN.BOOST.cut(y,X,sam=1:n,m.h,deg.h,nu.h,lambda.h,nu.f,M,family,monotone="TRUE")
m.opt.cut<-which.min(dev.vec.cut)
result <- list("MODEL"=optMOD, "DEV.VEC"=dev.vec, "M.OPT"=m.opt, "DEV.VEC.cut"=dev.vec.cut, "M.OPT.cut"=m.opt.cut)
return(result)
}

#########
#
# function for prediction 
#
###########

# obj: is an object of opt.model.cut or opt.model
# y.new: response of new data. if miss set 'NULL'
# X.new: standardized design matrix

predict.SI.LIN.BOOST.cut<-function(obj,y.new,X.new)
{
  a.new<-obj$Alpha
  b.new<-obj$Beta
  deg.h<-obj$Deg.h   
  eta.norm.new<-X.new%*%b.new
  KNOTS<-obj$Knots
  B.new<-bsplineS(eta.norm.new,breaks=KNOTS, norder=deg.h+1, nderiv=0)
  mu.hat<-family$linkinv(B.new%*%a.new)
  DEV<-'NULL'
  if(all(y.new) !='NULL')
  { 
  DEV<-sum(family$dev.resids(y.new,mu.hat,rep(1,nrow(X.new))))
  }
  result<-list("mu.hat"=mu.hat,"deviance"=DEV)
  return(result)
}

#########
#
# function for selecting the optimal "cut" model 
#
###########
opt.model.cut<-function(obj)
{
result<-list("Eta"=obj$MODEL$Eta.iter.norm[,obj$M.OPT.cut],
                 "Alpha"=obj$MODEL$Alpha.norm[,obj$M.OPT.cut],
                 "Beta"=obj$MODEL$Beta.cut[,obj$M.OPT.cut],
                 "Knots"=obj$MODEL$knots.h,
                 "Mu.hat"=obj$MODEL$MU.hat[,obj$M.OPT.cut],
                 "Lambda"=lambda.h,
                 "M.OPT"=obj$M.OPT.cut,
                 "Deg.h"=obj$MODEL$Deg.h)
return(result)
}
#########
#
# function for selecting the optimal model 
#
###########

opt.model<-function(obj)
{
result<-list("Eta"=obj$MODEL$Eta.iter.norm[,obj$M.OPT],
                 "Alpha"=obj$MODEL$Alpha.norm[,obj$M.OPT],
                 "Beta"=obj$MODEL$Beta[,obj$M.OPT],
                 "Knots"=obj$MODEL$knots.h,
                 "Mu.hat"=obj$MODEL$MU.hat[,obj$M.OPT],
                 "Lambda"=lambda.h,
                 "M.OPT"=obj$M.OPT,
                 "Deg.h"=obj$MODEL$Deg.h)
return(result)
}
